import numpy as np
import pandas as pd
from copy import deepcopy

def main(net):
  '''
  calculate pvalue of category closeness
  '''
  # calculate the distance between the data points within the same category and
  # compare to null distribution
  for inst_rc in ['row', 'col']:

    inst_nodes = deepcopy(net.dat['nodes'][inst_rc])

    inst_index = deepcopy(net.dat['node_info'][inst_rc]['clust'])

    # reorder based on clustered order
    inst_nodes = [ inst_nodes[i] for i in inst_index]

    # make distance matrix dataframe
    dm = dist_matrix_lattice(inst_nodes)

    node_infos = list(net.dat['node_info'][inst_rc].keys())

    all_cats = []
    for inst_info in node_infos:
      if 'dict_cat_' in inst_info:
        all_cats.append(inst_info)

    for cat_dict in all_cats:

      tmp_dict = net.dat['node_info'][inst_rc][cat_dict]

      pval_name = cat_dict.replace('dict_','pval_')
      net.dat['node_info'][inst_rc][pval_name] = {}

      for cat_name in tmp_dict:

        subset = tmp_dict[cat_name]

        inst_median = calc_median_dist_subset(dm, subset)

        hist = calc_hist_distances(dm, subset, inst_nodes)

        pval = 0

        for i in range(len(hist['prob'])):
          if i == 0:
            pval = hist['prob'][i]
          if i >= 1:
            if inst_median >= hist['bins'][i]:
              pval = pval + hist['prob'][i]

        net.dat['node_info'][inst_rc][pval_name][cat_name] = pval

def dist_matrix_lattice(names):
  from scipy.spatial.distance import pdist, squareform

  lattice_size = len(names)
  mat = np.zeros([lattice_size, 1])
  mat[:,0] = list(range(lattice_size))

  inst_dm = pdist(mat, metric='euclidean')

  inst_dm[inst_dm < 0] = float(0)

  inst_dm = squareform(inst_dm)

  df = pd.DataFrame(data=inst_dm, columns=names, index=names)

  return df


def calc_median_dist_subset(dm, subset):
  return np.median(dm[subset].loc[subset].values)

def calc_hist_distances(dm, subset, inst_nodes):
  np.random.seed(100)

  num_null = 1000
  num_points = len(subset)

  median_dist = []
  for i in range(num_null):
    tmp = np.random.choice(inst_nodes, num_points, replace=False)
    median_dist.append( np.median(dm[tmp].loc[tmp].values)  )

  tmp_dist = sorted(deepcopy(median_dist))

  median_dist = np.asarray(median_dist)
  s1 = pd.Series(median_dist)
  hist = np.histogram(s1, bins=30)

  H = {}
  H['prob'] = hist[0]/np.float(num_null)
  H['bins'] = hist[1]

  return H


def cluster_row_and_col(net, dist_type='cosine', linkage_type='average',
                        dendro=True, run_clustering=True, run_rank=True,
                        ignore_cat=False, calc_cat_pval=False, links=False,
                        clust_library='scipy', min_samples=1, min_cluster_size=2):
  ''' cluster net.dat and make visualization json, net.viz.
  optionally leave out dendrogram colorbar groups with dendro argument '''

  # import umap
  import scipy
  from copy import deepcopy
  from scipy.spatial.distance import pdist
  from . import categories, make_viz, cat_pval

  dm = {}
  for axis in ['row', 'col']:

    # save directly to dat structure
    node_info = net.dat['node_info'][axis]

    node_info['ini'] = list(range( len(net.dat['nodes'][axis]), -1, -1))

    tmp_mat = deepcopy(net.dat['mat'])

    # calc distance matrix
    if clust_library != 'hdbscan':
      dm[axis] = calc_distance_matrix(tmp_mat, axis, dist_type)
    else:
      dm[axis] = None

    # dm[axis] = calc_distance_matrix(tmp_mat, axis, dist_type)

    # cluster
    if run_clustering is True:
      node_info['clust'], node_info['Y'] = clust_and_group(net,
                                                     dm[axis],
                                                     axis,
                                                     tmp_mat,
                                                     dist_type=dist_type,
                                                     linkage_type=linkage_type,
                                                     clust_library=clust_library,
                                                     min_samples=min_samples,
                                                     min_cluster_size=min_cluster_size)
    else:
      dendro = False
      node_info['clust'] = node_info['ini']

    # sorting
    if run_rank is True:
      node_info['rank'] = sort_rank_nodes(net, axis, 'sum')
      node_info['rankvar'] = sort_rank_nodes(net, axis, 'var')
    else:
      node_info['rank'] = node_info['ini']
      node_info['rankvar'] = node_info['ini']

    ##################################
    if ignore_cat is False:
      categories.calc_cat_clust_order(net, axis)

  if calc_cat_pval is True:
    cat_pval.main(net)

  # make the visualization json
  make_viz.viz_json(net, dendro, links)

  return dm

def calc_distance_matrix(tmp_mat, axis, dist_type='cosine'):
  from scipy.spatial.distance import pdist
  import numpy as np

  if axis == 'row':
    inst_dm = pdist(tmp_mat, metric=dist_type)
  elif axis == 'col':
    inst_dm = pdist(tmp_mat.transpose(), metric=dist_type)

  inst_dm[inst_dm < 0] = float(0)

  return inst_dm

def clust_and_group(net, inst_dm, axis, mat, dist_type='cosine', linkage_type='average',
                    clust_library='scipy', min_samples=1, min_cluster_size=2):

  # print(clust_library)

  import scipy.cluster.hierarchy as hier
  import pandas as pd

  if clust_library == 'scipy':
    Y = hier.linkage(inst_dm, method=linkage_type)

  elif clust_library == 'fastcluster':
    import fastcluster
    Y = fastcluster.linkage(inst_dm, method=linkage_type)

  elif clust_library == 'hdbscan':
    # print('HDBSCAN!')
    import hdbscan


    # pca-umap-hdbscan using data (no pre-cal distance matrix)
    ######################################################
    from sklearn.decomposition import PCA
    clusterer = hdbscan.HDBSCAN(min_samples=min_samples,
                                min_cluster_size=min_cluster_size)

    # rows are the data points, cols are dimensions
    n_components = 50
    if axis == 'row':
      if mat.shape[1] > n_components:
        low_d_mat = PCA(n_components=n_components).fit_transform(mat)
      else:
        low_d_mat = mat

    elif axis == 'col':
      if mat.shape[0] > n_components:
        low_d_mat = PCA(n_components=n_components).fit_transform(mat.transpose())
      else:
        low_d_mat = mat.transpose()


    # run UMAP on low_d_mat (after PCA)
    # print('running umap!!!!!!!!!!!!!!!!!!!!!!!!!!')
    import umap
    umap_mat = umap.UMAP(
                          metric=dist_type,
                          n_neighbors=5,
                          min_dist=0.0,
                          n_components=2,
                          random_state=42,
                          ).fit_transform(low_d_mat)

    umap_df = pd.DataFrame(umap_mat.transpose(),
                           index=['x','y'],
                           columns=net.dat['nodes'][axis])

    net.umap[axis] = umap_df
    clusterer.fit(umap_mat)

    Y = clusterer.single_linkage_tree_.to_numpy()
    Z = hier.dendrogram(Y, no_plot=True)

  Z = hier.dendrogram(Y, no_plot=True)
  inst_clust_order = Z['leaves']

  return inst_clust_order, Y

def sort_rank_nodes(net, rowcol, rank_type):
  import numpy as np
  from operator import itemgetter
  from copy import deepcopy

  tmp_nodes = deepcopy(net.dat['nodes'][rowcol])
  inst_mat = deepcopy(net.dat['mat'])

  sum_term = []
  for i in range(len(tmp_nodes)):
    inst_dict = {}
    inst_dict['name'] = tmp_nodes[i]

    if rowcol == 'row':
      if rank_type == 'sum':
        inst_dict['rank'] = np.sum(inst_mat[i, :])
      elif rank_type == 'var':
        inst_dict['rank'] = np.var(inst_mat[i, :])
    else:
      if rank_type == 'sum':
        inst_dict['rank'] = np.sum(inst_mat[:, i])
      elif rank_type == 'var':
        inst_dict['rank'] = np.var(inst_mat[:, i])

    sum_term.append(inst_dict)

  sum_term = sorted(sum_term, key=itemgetter('rank'), reverse=False)

  tmp_sort_nodes = []
  for inst_dict in sum_term:
    tmp_sort_nodes.append(inst_dict['name'])

  sort_index = []
  for inst_node in tmp_nodes:
    sort_index.append(tmp_sort_nodes.index(inst_node))

  return sort_index



def main(self, widget=None):

  if hasattr(self, 'meta_cat') == False:
    self.meta_cat = False

  self.dat = {}
  self.dat['nodes'] = {}
  self.dat['nodes']['row'] = []
  self.dat['nodes']['col'] = []
  self.dat['mat'] = []

  self.dat['node_info'] = {}
  for inst_rc in self.dat['nodes']:
    self.dat['node_info'][inst_rc] = {}
    self.dat['node_info'][inst_rc]['ini'] = []
    self.dat['node_info'][inst_rc]['clust'] = []
    self.dat['node_info'][inst_rc]['rank'] = []
    self.dat['node_info'][inst_rc]['info'] = []
    self.dat['node_info'][inst_rc]['cat'] = []
    self.dat['node_info'][inst_rc]['value'] = []

  # check if net has categories predefined
  if hasattr(self, 'persistent_cat_colors') == False:
    self.persistent_cat_colors = False
    found_cats = False
  else:
    found_cats = True
    inst_cat_colors = self.viz['cat_colors']
    inst_global_cat_colors = self.viz['global_cat_colors']

  # initialize matrix colors
  ###########################
  has_matrix_colors = False
  if hasattr(self, 'viz'):
    if 'matrix_colors' in self.viz:
      has_matrix_colors = True

  if has_matrix_colors:
    matrix_colors = self.viz['matrix_colors']
  else:
    matrix_colors = {}
    matrix_colors['pos'] = 'red'
    matrix_colors['neg'] = 'blue'

  # add widget if necessary
  if widget != None:
    self.widget_class = widget

  self.is_downsampled = False

  self.viz = {}
  self.viz['row_nodes'] = []
  self.viz['col_nodes'] = []
  self.viz['links'] = []
  self.viz['mat'] = []

  if found_cats == False:
    self.viz['cat_colors'] = {}
    self.viz['cat_colors']['row'] = {}
    self.viz['cat_colors']['col'] = {}
    self.viz['global_cat_colors'] = {}
  else:
    self.viz['cat_colors'] = inst_cat_colors
    self.viz['global_cat_colors'] = inst_global_cat_colors

  self.viz['matrix_colors'] = matrix_colors

  self.sim = {}

  self.umap = {}


def viz(self, reset_cat_colors=False):

  # keep track of old cat_colors
  old_cat_colors = self.viz['cat_colors']
  old_global_cat_colors = self.viz['global_cat_colors']

  if 'matrix_colors' in self.viz:
    matrix_colors = self.viz['matrix_colors']

  self.viz = {}
  self.viz['row_nodes'] = []
  self.viz['col_nodes'] = []
  self.viz['links'] = []
  self.viz['mat'] = []

  if reset_cat_colors == True:
    self.viz['cat_colors'] = {}
    self.viz['cat_colors']['row'] = {}
    self.viz['cat_colors']['col'] = {}
    self.viz['global_cat_colors'] = {}
  else:
    self.viz['cat_colors'] = old_cat_colors
    self.viz['global_cat_colors'] = old_global_cat_colors

  self.viz['matrix_colors'] = matrix_colors



"""Top-level package for clustergrammer2_core."""

from .clustergrammer_fun import *

net = Network()

__author__ = """Nicolas Fernandez"""
__email__ = 'nickfloresfernandez@gmail.com'
__version__ = '0.1.0'
